from typing import Literal

from qtpy import QtMultimedia

from prettyqt import core, multimedia
from prettyqt.utils import InvalidParamError, bidict


COLOR_FILTERS = bidict(
    none=QtMultimedia.QCameraImageProcessing.ColorFilterNone,
    grayscale=QtMultimedia.QCameraImageProcessing.ColorFilterGrayscale,
    negative=QtMultimedia.QCameraImageProcessing.ColorFilterNegative,
    solarize=QtMultimedia.QCameraImageProcessing.ColorFilterSolarize,
    sepia=QtMultimedia.QCameraImageProcessing.ColorFilterSepia,
    posterize=QtMultimedia.QCameraImageProcessing.ColorFilterPosterize,
    whiteboard=QtMultimedia.QCameraImageProcessing.ColorFilterWhiteboard,
    blackboard=QtMultimedia.QCameraImageProcessing.ColorFilterBlackboard,
    aqua=QtMultimedia.QCameraImageProcessing.ColorFilterAqua,
    vendor=QtMultimedia.QCameraImageProcessing.ColorFilterVendor,
)

ColorFilterStr = Literal[
    "none",
    "grayscale",
    "negative",
    "solarize",
    "sepia",
    "posterize",
    "whiteboard",
    "blackboard",
    "aqua",
    "vendor",
]

WHITE_BALANCE_MODE = bidict(
    auto=QtMultimedia.QCameraImageProcessing.WhiteBalanceAuto,
    manual=QtMultimedia.QCameraImageProcessing.WhiteBalanceManual,
    sunlight=QtMultimedia.QCameraImageProcessing.WhiteBalanceSunlight,
    cloudy=QtMultimedia.QCameraImageProcessing.WhiteBalanceCloudy,
    shade=QtMultimedia.QCameraImageProcessing.WhiteBalanceShade,
    tungsten=QtMultimedia.QCameraImageProcessing.WhiteBalanceTungsten,
    fluorescent=QtMultimedia.QCameraImageProcessing.WhiteBalanceFluorescent,
    flash=QtMultimedia.QCameraImageProcessing.WhiteBalanceFlash,
    sunset=QtMultimedia.QCameraImageProcessing.WhiteBalanceSunset,
    vendor=QtMultimedia.QCameraImageProcessing.WhiteBalanceVendor,
)

WhiteBalanceModeStr = Literal[
    "auto",
    "manual",
    "sunlight",
    "cloudy",
    "shade",
    "tungsten",
    "fluorescent",
    "flash",
    "sunset",
    "vendor",
]

QtMultimedia.QCameraImageProcessing.__bases__ = (core.Object,)


class CameraImageProcessing(core.Object):
    def __init__(self, item: QtMultimedia.QCameraImageProcessing):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def set_color_filter(self, mode: ColorFilterStr):
        """Set the color filter.

        Args:
            mode: color filter

        Raises:
            InvalidParamError: color filter does not exist
        """
        if mode not in COLOR_FILTERS:
            raise InvalidParamError(mode, COLOR_FILTERS)
        self.item.setColorFilter(COLOR_FILTERS[mode])

    def get_color_filter(self) -> ColorFilterStr:
        """Return current exposure mode.

        Returns:
            color filter
        """
        return COLOR_FILTERS.inverse[self.item.colorFilter()]

    def set_white_balance_mode(self, mode: WhiteBalanceModeStr):
        """Set the white balance mode.

        Args:
            mode: white balance mode

        Raises:
            InvalidParamError: white balance mode does not exist
        """
        if mode not in WHITE_BALANCE_MODE:
            raise InvalidParamError(mode, WHITE_BALANCE_MODE)
        self.item.setWhiteBalanceMode(WHITE_BALANCE_MODE[mode])

    def get_white_balance_mode(self) -> WhiteBalanceModeStr:
        """Return current white balance mode.

        Returns:
            white balance mode
        """
        return WHITE_BALANCE_MODE.inverse[self.item.whiteBalanceMode()]

    def is_color_filter_supported(self, filter_: ColorFilterStr) -> bool:
        if filter_ not in COLOR_FILTERS:
            raise InvalidParamError(filter_, COLOR_FILTERS)
        return self.item.isColorFilterSupported(COLOR_FILTERS[filter_])

    def is_white_balance_mode_supported(self, mode: WhiteBalanceModeStr) -> bool:
        if mode not in WHITE_BALANCE_MODE:
            raise InvalidParamError(mode, WHITE_BALANCE_MODE)
        return self.item.isWhiteBalanceModeSupported(WHITE_BALANCE_MODE[mode])


if __name__ == "__main__":
    cam = multimedia.Camera()
    focus = cam.get_focus()



from qtpy import QtCore


class Singleton(type(QtCore.QObject), type):  # type: ignore
    def __init__(cls, name, bases, dict):
        super().__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kw)
        return cls.instance


if __name__ == "__main__":

    class Test(QtCore.QObject, metaclass=Singleton):
        pass



from typing import Union

from qtpy import QtCore, QtNetwork

from prettyqt import core
from prettyqt.utils import InvalidParamError, bidict


CONTENT_TYPES = bidict(
    mixed=QtNetwork.QHttpMultiPart.MixedType,
    related=QtNetwork.QHttpMultiPart.RelatedType,
    form=QtNetwork.QHttpMultiPart.FormDataType,
    alternative=QtNetwork.QHttpMultiPart.AlternativeType,
)

QtNetwork.QHttpMultiPart.__bases__ = (core.Object,)


class HttpMultiPart(QtNetwork.QHttpMultiPart):
    def __add__(self, other: QtNetwork.QHttpPart):
        self.append(other)
        return self

    def set_content_type(self, typ: str):
        """Set content type.

        Valid values: "mixed", "related", "form", "alternative"

        Args:
            typ: content type

        Raises:
            InvalidParamError: content type does not exist
        """
        if typ not in CONTENT_TYPES:
            raise InvalidParamError(typ, CONTENT_TYPES)
        self.setContentType(CONTENT_TYPES[typ])

    def set_boundary(self, boundary: Union[str, bytes, QtCore.QByteArray]):
        if isinstance(boundary, str):
            boundary = boundary.encode()
        if isinstance(boundary, bytes):
            boundary = QtCore.QByteArray(boundary)
        self.setBoundary(boundary)

    def get_boundary(self) -> str:
        return bytes(self.boundary()).decode()


if __name__ == "__main__":
    part = HttpMultiPart()



from typing import Tuple

from qtpy import QtGui

from prettyqt import core, gui


QtGui.QImage.__bases__ = (gui.PaintDevice,)


class Image(QtGui.QImage):
    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __setitem__(self, index: Tuple[int, int], value):
        self.setPixel(index[0], index[1], value)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __getitem__(self, index: Tuple[int, int]):
        return self.pixel(index[0], index[1])

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)


if __name__ == "__main__":
    image = Image()
    bytes(image)



from typing import Optional, Union

from qtpy import QtWidgets

from prettyqt.utils import bidict


ITERATOR_FLAGS = bidict(
    all=QtWidgets.QTreeWidgetItemIterator.All,
    hidden=QtWidgets.QTreeWidgetItemIterator.Hidden,
    not_hidden=QtWidgets.QTreeWidgetItemIterator.NotHidden,
    selected=QtWidgets.QTreeWidgetItemIterator.Selected,
    unselected=QtWidgets.QTreeWidgetItemIterator.Unselected,
    selectable=QtWidgets.QTreeWidgetItemIterator.Selectable,
    not_selectable=QtWidgets.QTreeWidgetItemIterator.NotSelectable,
    drag_enabled=QtWidgets.QTreeWidgetItemIterator.DragEnabled,
    drag_disabled=QtWidgets.QTreeWidgetItemIterator.DragDisabled,
    drop_enabled=QtWidgets.QTreeWidgetItemIterator.DropEnabled,
    drop_disabled=QtWidgets.QTreeWidgetItemIterator.DropDisabled,
    has_children=QtWidgets.QTreeWidgetItemIterator.HasChildren,
    no_children=QtWidgets.QTreeWidgetItemIterator.NoChildren,
    checked=QtWidgets.QTreeWidgetItemIterator.Checked,
    not_checked=QtWidgets.QTreeWidgetItemIterator.NotChecked,
    enabled=QtWidgets.QTreeWidgetItemIterator.Enabled,
    disabled=QtWidgets.QTreeWidgetItemIterator.Disabled,
    editable=QtWidgets.QTreeWidgetItemIterator.Editable,
    not_editable=QtWidgets.QTreeWidgetItemIterator.NotEditable,
    user_flag=QtWidgets.QTreeWidgetItemIterator.UserFlag,
)


class TreeWidgetItemIterator(QtWidgets.QTreeWidgetItemIterator):
    def __init__(
        self,
        other: Union[
            QtWidgets.QTreeWidget,
            QtWidgets.QTreeWidgetItem,
            QtWidgets.QTreeWidgetItemIterator,
        ],
        flags: Optional[QtWidgets.QTreeWidgetItemIterator.IteratorFlags] = None,
        hidden: Optional[bool] = None,
        selected: Optional[bool] = None,
        selectable: Optional[bool] = None,
        draggable: Optional[bool] = None,
        droppable: Optional[bool] = None,
        has_children: Optional[bool] = None,
        checked: Optional[bool] = None,
        enabled: Optional[bool] = None,
        editable: Optional[bool] = None,
        user_flag: bool = False,
    ):
        if isinstance(other, QtWidgets.QTreeWidgetItemIterator):
            super().__init__(other)
        else:
            if flags is None:
                flags = QtWidgets.QTreeWidgetItemIterator.All
            if hidden is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.Hidden
            elif hidden is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.NotHidden
            if selected is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.Selected
            elif selected is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.Unselected
            if selectable is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.Selectable
            elif selectable is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.NotSelectable
            if draggable is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.DragEnabled
            elif draggable is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.DragDisabled
            if droppable is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.DropEnabled
            elif droppable is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.DropDisabled
            if has_children is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.HasChildren
            elif has_children is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.NoChildren
            if checked is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.Checked
            elif checked is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.NotChecked
            if enabled is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.Enabled
            elif enabled is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.Disabled
            if editable is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.Editable
            elif editable is False:
                flags |= QtWidgets.QTreeWidgetItemIterator.NotEditable
            if user_flag is True:
                flags |= QtWidgets.QTreeWidgetItemIterator.UserFlag
            super().__init__(other, flags)  # type: ignore


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    item = QtWidgets.QTreeWidget()
    widget = TreeWidgetItemIterator(item)



from typing import Literal

from qtpy import PYQT5, PYSIDE2


if PYQT5:
    from PyQt5 import QtPositioning
elif PYSIDE2:
    from PySide2 import QtPositioning

from prettyqt import core, positioning
from prettyqt.utils import bidict


ATTRIBUTE = bidict(
    direction=QtPositioning.QGeoPositionInfo.Direction,
    ground_speed=QtPositioning.QGeoPositionInfo.GroundSpeed,
    vertical_speed=QtPositioning.QGeoPositionInfo.VerticalSpeed,
    magnetic_variation=QtPositioning.QGeoPositionInfo.MagneticVariation,
    horizontal_accuracy=QtPositioning.QGeoPositionInfo.HorizontalAccuracy,
    vertical_accuracy=QtPositioning.QGeoPositionInfo.VerticalAccuracy,
)

AttributeStr = Literal[
    "direction",
    "ground_speed",
    "vertical_speed",
    "magnetic_variation",
    "horizontal_accuracy",
    "vertical_accuracy",
]


class GeoPositionInfo(QtPositioning.QGeoPositionInfo):
    def __repr__(self):
        return f"{type(self).__name__}({self.get_coordinate()}, {self.get_timestamp()})"

    def __contains__(self, index: AttributeStr):
        return self.hasAttribute(ATTRIBUTE[index])

    def __getitem__(self, index: AttributeStr) -> float:
        return self.attribute(ATTRIBUTE[index])

    def __setitem__(self, index: AttributeStr, value: float):
        self.setAttribute(ATTRIBUTE[index], value)

    def __delitem__(self, index: AttributeStr):
        return self.removeAttribute(ATTRIBUTE[index])

    def __bool__(self):
        return self.isValid()

    def get_coordinate(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.coordinate())

    def get_timestamp(self) -> core.DateTime:
        return core.DateTime(self.timestamp())



from qtpy import QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict


CHECK_TYPE = bidict(
    not_checkable=QtWidgets.QStyleOptionMenuItem.NotCheckable,
    exclusive=QtWidgets.QStyleOptionMenuItem.Exclusive,
    non_exclusive=QtWidgets.QStyleOptionMenuItem.NonExclusive,
)

MENU_ITEM_TYPE = bidict(
    normal=QtWidgets.QStyleOptionMenuItem.Normal,
    default_item=QtWidgets.QStyleOptionMenuItem.DefaultItem,
    separator=QtWidgets.QStyleOptionMenuItem.Separator,
    sub_menu=QtWidgets.QStyleOptionMenuItem.SubMenu,
    scroller=QtWidgets.QStyleOptionMenuItem.Scroller,
    tear_off=QtWidgets.QStyleOptionMenuItem.TearOff,
    margin=QtWidgets.QStyleOptionMenuItem.Margin,
    empty_area=QtWidgets.QStyleOptionMenuItem.EmptyArea,
)


QtWidgets.QStyleOptionMenuItem.__bases__ = (widgets.StyleOption,)


class StyleOptionMenuItem(QtWidgets.QStyleOptionMenuItem):
    pass



from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QStyleOptionComboBox.__bases__ = (widgets.StyleOptionComplex,)


class StyleOptionComboBox(QtWidgets.QStyleOptionComboBox):
    pass



from prettyqt import gui


FORMAT_1 = gui.TextCharFormat()
FORMAT_1.set_background_color("lightgreen")
FORMAT_2 = gui.TextCharFormat()
FORMAT_2.set_background_color("lightblue")


class RegexMatchHighlighter(gui.SyntaxHighlighter):
    def __init__(self, document=None):
        super().__init__(document)
        self.spans = []

    def set_spans(self, spans):
        self.spans = spans
        # print(self.spans)
        self.rehighlight()

    def _colorize(self, line_pos, match_len, match_num):
        fmt = FORMAT_1 if match_num % 2 == 0 else FORMAT_2
        self.setFormat(line_pos, match_len, fmt)

    def highlightBlock(self, text: str):
        block = self.currentBlock()
        # line_no = block.blockNumber()
        # if line_no == 0:
        #     self.setCurrentBlockState(-1)
        start_char = block.position()
        end_char = start_char + block.length()
        # print(f"\nline {line_no} ({start_char} - {end_char})")
        # print(f"prev block state: {self.previousBlockState()}")
        if not self.spans or not text:
            return None
        for i, (start, end) in enumerate(self.spans):
            if end < start_char:
                continue
            if start > end_char:
                break
            starts_in_line = start_char <= start <= end_char
            ends_in_line = start_char <= end <= end_char
            if starts_in_line and ends_in_line:
                # print(f"in line: {line_pos} - {line_pos + match_len}")
                self._colorize(start - start_char, end - start, i)
            elif ends_in_line:
                # if self.previousBlockState() == 1:
                # print(f"ends: {end}")
                self._colorize(0, end - start, i)
                # self.setCurrentBlockState(-1)
            elif starts_in_line:
                # print(f"starts: {line_pos}")
                # self.setCurrentBlockState(1)
                self._colorize(start - start_char, end - start, i)



import contextlib
from typing import Literal

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import InvalidParamError, mappers


OPEN_MODES = mappers.FlagMap(
    QtCore.QIODevice.OpenModeFlag,
    not_open=QtCore.QIODevice.NotOpen,
    read_only=QtCore.QIODevice.ReadOnly,
    write_only=QtCore.QIODevice.WriteOnly,
    read_write=QtCore.QIODevice.ReadWrite,
    append=QtCore.QIODevice.Append,
    truncate=QtCore.QIODevice.Truncate,
    text=QtCore.QIODevice.Text,
    unbuffered=QtCore.QIODevice.Unbuffered,
    new_only=QtCore.QIODevice.NewOnly,
    existing_only=QtCore.QIODevice.ExistingOnly,
)

OpenModeStr = Literal[
    "not_open",
    "read_only",
    "write_only",
    "read_write",
    "append",
    "truncate",
    "text",
    "unbuffered",
    "new_only",
    "existing_only",
]

QtCore.QIODevice.__bases__ = (core.Object,)


class IODevice(QtCore.QIODevice):
    @contextlib.contextmanager
    def open_file(self, mode: OpenModeStr):
        if mode not in OPEN_MODES:
            raise InvalidParamError(mode, OPEN_MODES)
        self.open(OPEN_MODES[mode])
        yield self
        self.close()

    def get_open_mode(self) -> OpenModeStr:
        return OPEN_MODES.inverse[self.openMode()]



import contextlib
from typing import List

from qtpy import QtCore, QtGui

from prettyqt import constants, core, gui
from prettyqt.utils import InvalidParamError, colors


QtGui.QGuiApplication.__bases__ = (core.CoreApplication,)


class GuiApplication(QtGui.QGuiApplication):
    @classmethod
    @contextlib.contextmanager
    def override_cursor(cls, cursor: constants.CursorShapeStr):
        cls.set_override_cursor(cursor)
        yield cursor
        cls.restore_override_cursor()

    @classmethod
    def set_override_cursor(cls, cursor: constants.CursorShapeStr):
        cursor = gui.Cursor(constants.CURSOR_SHAPE[cursor])
        cls.setOverrideCursor(cursor)

    @classmethod
    def restore_override_cursor(cls):
        cls.restoreOverrideCursor()

    @classmethod
    def get_clipboard(cls) -> gui.Clipboard:
        return gui.Clipboard(cls.clipboard())

    def set_layout_direction(self, direction: constants.LayoutDirectionStr):
        """Set layout direction.

        Args:
            direction: layout direction

        Raises:
            InvalidParamError: layout direction does not exist
        """
        if direction not in constants.LAYOUT_DIRECTION:
            raise InvalidParamError(direction, constants.LAYOUT_DIRECTION)
        self.setLayoutDirection(constants.LAYOUT_DIRECTION[direction])

    def get_layout_direction(self) -> constants.LayoutDirectionStr:
        """Get the current layout direction.

        Returns:
            layout direction
        """
        return constants.LAYOUT_DIRECTION.inverse[self.layoutDirection()]

    @classmethod
    def set_high_dpi_scale_factor_rounding_policy(
        cls, policy: constants.HighDpiScaleFactorRoundingPolicyStr
    ):
        """Set high dpi scale factor rounding policy.

        Args:
            direction: rounding policy

        Raises:
            InvalidParamError: rounding policy does not exist
        """
        if policy not in constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY:
            raise InvalidParamError(
                policy, constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY
            )
        cls.setHighDpiScaleFactorRoundingPolicy(
            constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY[policy]
        )

    @classmethod
    def get_high_dpi_scale_factor_rounding_policy(
        cls,
    ) -> constants.HighDpiScaleFactorRoundingPolicyStr:
        """Get the current high dpi scale factor rounding policy.

        Returns:
            rounding policy
        """
        return constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY.inverse[
            cls.highDpiScaleFactorRoundingPolicy()
        ]

    @classmethod
    def get_application_state(cls) -> List[constants.ApplicationStateStr]:
        """Get the current application state.

        Returns:
            application state
        """
        return [
            k
            for k, v in constants.APPLICATION_STATES.items()
            if v & cls.applicationState()  # type: ignore
        ]

    def get_primary_screen(self) -> gui.Screen:
        return gui.Screen(self.primaryScreen())

    def get_screen_at(self, point: QtCore.QPoint) -> gui.Screen:
        return gui.Screen(self.screenAt(point))

    def get_screens(self) -> List[gui.Screen]:
        return [gui.Screen(i) for i in self.screens()]

    @classmethod
    def get_input_method(cls) -> gui.InputMethod:
        return gui.InputMethod(cls.inputMethod())

    @classmethod
    def copy_to_clipboard(cls, text: str):
        """Sets clipboard to supplied text."""
        cb = cls.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(text, mode=cb.Clipboard)

    @classmethod
    def get_font(cls) -> gui.Font:
        return gui.Font(cls.font())

    def set_icon(self, icon: gui.icon.IconType):
        """Set the default window icon.

        Args:
            icon: icon to use
        """
        icon = gui.icon.get_icon(icon, color=colors.WINDOW_ICON_COLOR)
        self.setWindowIcon(icon)

    def get_icon(self) -> gui.Icon:
        return gui.Icon(self.windowIcon())

    @classmethod
    def get_palette(cls) -> gui.Palette:
        return gui.Palette(cls.palette())



from typing import Literal

from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import InvalidParamError, bidict


MODES = bidict(
    standard=QtGui.QPageLayout.StandardMode, full_page=QtGui.QPageLayout.FullPageMode
)

ModeStr = Literal["standard", "full_page"]

ORIENTATIONS = bidict(
    portrait=QtGui.QPageLayout.Portrait, landscape=QtGui.QPageLayout.Landscape
)

OrientationStr = Literal["portrait", "landscape"]

UNITS = bidict(
    millimeter=QtGui.QPageLayout.Millimeter,
    point=QtGui.QPageLayout.Point,
    inch=QtGui.QPageLayout.Inch,
    pica=QtGui.QPageLayout.Pica,
    didot=QtGui.QPageLayout.Didot,
    cicero=QtGui.QPageLayout.Cicero,
)

UnitStr = Literal["millimeter", "point", "inch", "pica", "didot", "cicero"]


class PageLayout(QtGui.QPageLayout):
    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def serialize_fields(self):
        return dict(
            margins=self.margins(),
            minimum_margins=self.minimumMargins(),
            mode=self.get_mode(),
            orientation=self.get_orientation(),
            page_size=self.get_page_size(),
            units=self.get_units(),
        )

    def set_units(self, unit: UnitStr):
        """Set unit.

        Args:
            unit: unit

        Raises:
            InvalidParamError: unit does not exist
        """
        if unit not in UNITS:
            raise InvalidParamError(unit, UNITS)
        self.setUnits(UNITS[unit])

    def get_units(self) -> UnitStr:
        """Get the current unit.

        Returns:
            unit
        """
        return UNITS.inverse[self.units()]

    def set_mode(self, mode: ModeStr):
        """Set mode.

        Args:
            mode: mode

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        self.setMode(MODES[mode])

    def get_mode(self) -> ModeStr:
        """Get the current mode.

        Returns:
            mode
        """
        return MODES.inverse[self.mode()]

    def set_orientation(self, orientation: OrientationStr):
        """Set orientation.

        Args:
            orientation: orientation

        Raises:
            InvalidParamError: orientation does not exist
        """
        if orientation not in ORIENTATIONS:
            raise InvalidParamError(orientation, ORIENTATIONS)
        self.setOrientation(ORIENTATIONS[orientation])

    def get_orientation(self) -> OrientationStr:
        """Get the current orientation.

        Returns:
            orientation
        """
        return ORIENTATIONS.inverse[self.orientation()]

    def get_page_size(self) -> gui.PageSize:
        return gui.PageSize(self.pageSize())


if __name__ == "__main__":
    movie = PageLayout()



from qtpy.QtCharts import QtCharts

from prettyqt import charts


QtCharts.QHorizontalStackedBarSeries.__bases__ = (charts.AbstractBarSeries,)


class HorizontalStackedBarSeries(QtCharts.QHorizontalStackedBarSeries):
    pass



from typing import Literal

from deprecated import deprecated
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCharts import QtCharts

from prettyqt import charts, core, gui, widgets
from prettyqt.utils import InvalidParamError, bidict


RUBBER_BAND = bidict(
    none=QtCharts.QChartView.NoRubberBand,
    vertical=QtCharts.QChartView.VerticalRubberBand,
    horizontal=QtCharts.QChartView.HorizontalRubberBand,
    rectangle=QtCharts.QChartView.RectangleRubberBand,
)

RubberBandStr = Literal["none", "vertical", "horizontal", "rectangle"]


ZOOM_IN_FACTOR = 1.1
ZOOM_OUT_FACTOR = 1.0 / ZOOM_IN_FACTOR
SCROLL_STEP_SIZE = 10


class ChartView(QtCharts.QChartView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        chart = charts.Chart()
        self.setChart(chart)
        self.setRenderHint(gui.Painter.Antialiasing)
        self.set_rubber_band("rectangle")
        # self.setDragMode(self.RubberBandDrag)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        """Handle keypress events to allow navigation via keyboard."""
        key = event.key()
        if key == QtCore.Qt.Key_Escape:
            self.chart().zoomReset()
        elif key == QtCore.Qt.Key_Plus:
            self.chart().zoom_by_factor(ZOOM_IN_FACTOR)
        elif key == QtCore.Qt.Key_Minus:
            self.chart().zoom_by_factor(ZOOM_OUT_FACTOR)
        elif key == QtCore.Qt.Key_Left:
            self.chart().scroll(-SCROLL_STEP_SIZE, 0)
        elif key == QtCore.Qt.Key_Right:
            self.chart().scroll(SCROLL_STEP_SIZE, 0)
        elif key == QtCore.Qt.Key_Up:
            self.chart().scroll(0, SCROLL_STEP_SIZE)
        elif key == QtCore.Qt.Key_Down:
            self.chart().scroll(0, -SCROLL_STEP_SIZE)
        elif key == QtCore.Qt.Key_0:
            self.chart().apply_nice_numbers()
        else:
            return super().keyPressEvent(event)
        event.accept()

    def wheelEvent(self, event: QtGui.QWheelEvent):
        """Handle wheel event for zooming."""
        fct = ZOOM_IN_FACTOR if event.angleDelta().y() > 0 else ZOOM_OUT_FACTOR
        self.chart().zoom_by_factor(fct)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        """Override to allow dragging the chart."""
        if event.button() == QtCore.Qt.RightButton:
            widgets.Application.restoreOverrideCursor()
            event.accept()
            return None
        super().mouseReleaseEvent(event)

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        """Override to allow dragging the chart."""
        if event.button() == QtCore.Qt.RightButton:
            cursor = gui.Cursor(QtCore.Qt.SizeAllCursor)
            widgets.Application.setOverrideCursor(cursor)
            self.last_mouse_pos = event.pos()
            event.accept()

        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        """Override to allow dragging the chart."""
        # pan the chart with a middle mouse drag
        if event.buttons() & QtCore.Qt.RightButton:
            if not self.last_mouse_pos:
                return None
            pos_diff = event.pos() - self.last_mouse_pos
            self.chart().scroll(-pos_diff.x(), pos_diff.y())

            self.last_mouse_pos = event.pos()
            event.accept()

            widgets.Application.restoreOverrideCursor()

        super().mouseMoveEvent(event)

    @deprecated(reason="This method is deprecated, use save_as_image instead.")
    def save(self):
        self.save_as_image()

    @core.Slot()
    def save_as_image(self):
        """Let user choose folder and save chart as an image file."""
        dlg = widgets.FileDialog(mode="save", caption="Save image")
        filters = {"Bmp files": [".bmp"], "Jpeg files": [".jpg"], "Png files": [".png"]}
        dlg.set_extension_filter(filters)
        filename = dlg.open_file()
        if not filename:
            return None
        self.chart().show_legend()
        image = self.get_image()
        image.save(str(filename[0]))
        self.chart().hide_legend()

    def get_image(self) -> QtGui.QPixmap:
        image = self.grab()
        gl_widget = self.findChild(QtWidgets.QOpenGLWidget)
        if gl_widget:
            d = gl_widget.mapToGlobal(core.Point()) - self.mapToGlobal(core.Point())
            with gui.Painter(image) as painter:
                painter.set_composition_mode("source_atop")
                painter.drawImage(d, gl_widget.grabFramebuffer())
        return image

    def set_rubber_band(self, typ: RubberBandStr):
        """Set the rubber band type.

        Args:
            typ: rubber band type

        Raises:
            InvalidParamError: rubber band type does not exist
        """
        if typ not in RUBBER_BAND:
            raise InvalidParamError(typ, RUBBER_BAND)
        self.setRubberBand(RUBBER_BAND[typ])

    def get_rubber_band(self) -> RubberBandStr:
        """Return current rubber band type.

        Returns:
            Rubber band type
        """
        return RUBBER_BAND.inverse[self.rubberBand()]

    # def hide_legend(self):
    #     self.chart().hide_legend()

    # def show_legend(self):
    #     self.chart().show_legend()

    # def set_legend_alignment(self, alignment: str):
    #     if alignment not in constants.SIDES:
    #         raise ValueError(f"{alignment!r} not a valid alignment.")
    #     self.chart().legend().setAlignment(constants.SIDES[alignment])



from typing import Literal

from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import InvalidParamError, bidict, prettyprinter


CAPITALIZATION = bidict(
    mixed_case=QtGui.QFont.MixedCase,
    all_uppercase=QtGui.QFont.AllUppercase,
    all_lowercase=QtGui.QFont.AllLowercase,
    small_caps=QtGui.QFont.SmallCaps,
    capitalize=QtGui.QFont.Capitalize,
)

CapitalizationStr = Literal[
    "mixed_case", "all_uppercase", "all_lowercase", "small_caps", "capitalize"
]

HINTING_PREFERENCE = bidict(
    default=QtGui.QFont.PreferDefaultHinting,
    none=QtGui.QFont.PreferNoHinting,
    vertical=QtGui.QFont.PreferVerticalHinting,
    full=QtGui.QFont.PreferFullHinting,
)

HintingPreferenceStr = Literal["default", "none", "vertical", "full"]

SPACING_TYPE = bidict(
    percentage=QtGui.QFont.PercentageSpacing,
    absolute=QtGui.QFont.AbsoluteSpacing,
)

SpacingTypeStr = Literal["percentage", "absolute"]

STRETCH = bidict(
    any=QtGui.QFont.AnyStretch,
    ultra_condensed=QtGui.QFont.UltraCondensed,
    extra_condensed=QtGui.QFont.ExtraCondensed,
    condensed=QtGui.QFont.Condensed,
    semi_condensed=QtGui.QFont.SemiCondensed,
    unstretched=QtGui.QFont.Unstretched,
    semi_expanded=QtGui.QFont.SemiExpanded,
    expanded=QtGui.QFont.Expanded,
    extra_expanded=QtGui.QFont.ExtraExpanded,
    ultra_expanded=QtGui.QFont.UltraExpanded,
)

StretchStr = Literal[
    "any",
    "ultra_condensed",
    "extra_condensed",
    "condensed",
    "semi_condensed",
    "unstretched",
    "semi_expanded",
    "expanded",
    "extra_expanded",
    "ultra_expanded",
]

STYLE = bidict(
    normal=QtGui.QFont.StyleNormal,
    italic=QtGui.QFont.StyleItalic,
    oblique=QtGui.QFont.StyleOblique,
)

StyleStr = Literal["normal", "italic", "oblique"]

STYLE_STRATEGY = bidict(
    prefer_default=QtGui.QFont.PreferDefault,
    prefer_bitmap=QtGui.QFont.PreferBitmap,
    prefer_device=QtGui.QFont.PreferDevice,
    prefer_outline=QtGui.QFont.PreferOutline,
    force_outline=QtGui.QFont.ForceOutline,
    no_antialias=QtGui.QFont.NoAntialias,
    so_subpixel_antialias=QtGui.QFont.NoSubpixelAntialias,
    prefer_antialias=QtGui.QFont.PreferAntialias,
    open_gl_compatible=QtGui.QFont.OpenGLCompatible,
    no_font_merging=QtGui.QFont.NoFontMerging,
    prefer_no_shaping=QtGui.QFont.PreferNoShaping,
)  # ORed with PreferMatch, PreferQuality, ForceIntegerMetrics

StyleStrategyStr = Literal[
    "prefer_default",
    "prefer_bitmap",
    "prefer_device",
    "prefer_outline",
    "force_outline",
    "no_antialias",
    "so_subpixel_antialias",
    "prefer_antialias",
    "open_gl_compatible",
    "no_font_merging",
    "prefer_no_shaping",
]

STYLE_HINTS = bidict(
    any=QtGui.QFont.AnyStyle,
    sans_serif=QtGui.QFont.SansSerif,
    serif=QtGui.QFont.Serif,
    typewriter=QtGui.QFont.TypeWriter,
    decorative=QtGui.QFont.Decorative,
    monospace=QtGui.QFont.Monospace,
    fantasy=QtGui.QFont.Fantasy,
    cursive=QtGui.QFont.Cursive,
    system=QtGui.QFont.System,
)

StyleHintStr = Literal[
    "any",
    "sans_serif",
    "serif",
    "typewriter",
    "decorative",
    "monospace",
    "fantasy",
    "cursive",
    "system",
]

WEIGHT = bidict(
    thin=QtGui.QFont.Thin,
    extra_light=QtGui.QFont.ExtraLight,
    light=QtGui.QFont.Light,
    normal=QtGui.QFont.Normal,
    medium=QtGui.QFont.Medium,
    demi_bold=QtGui.QFont.DemiBold,
    bold=QtGui.QFont.Bold,
    extra_bold=QtGui.QFont.ExtraBold,
    black=QtGui.QFont.Black,
)

WeightStr = Literal[
    "thin",
    "extra_light",
    "light",
    "normal",
    "medium",
    "demi_bold",
    "bold",
    "extra_bold",
    "black",
]


class Font(prettyprinter.PrettyPrinter, QtGui.QFont):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self.family()!r}, {self.pointSize()}, "
            f"{self.weight()}, {self.italic()})"
        )

    def __getstate__(self):
        return dict(
            family=self.family(),
            pointsize=self.pointSize(),
            weight=self.weight(),
            italic=self.italic(),
        )

    def __setstate__(self, state):
        self.setFamily(state["family"])
        if state["pointsize"] > -1:
            self.setPointSize(state["pointsize"])
        self.setWeight(state["weight"])
        self.setItalic(state["italic"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize(self):
        return self.__getstate__()

    @property
    def metrics(self):
        return gui.FontMetrics(self)

    def set_size(self, size: int):
        self.setPointSize(size)

    @classmethod
    def mono(cls, size=8):
        font = cls("Consolas", size)
        # font.setStyleHint()
        return font

    def set_style_hint(self, hint: StyleHintStr):
        """Set the style hint.

        Args:
            hint: style hint

        Raises:
            InvalidParamError: invalid style hint
        """
        if hint not in STYLE_HINTS:
            raise InvalidParamError(hint, STYLE_HINTS)
        self.setStyleHint(STYLE_HINTS[hint])

    def set_weight(self, weight: WeightStr):
        """Set the font weight.

        Args:
            weight: font weight

        Raises:
            InvalidParamError: invalid font weight
        """
        if weight not in WEIGHT:
            raise InvalidParamError(weight, WEIGHT)
        self.setWeight(WEIGHT[weight])

    def get_weight(self) -> WeightStr:
        """Get current font weight.

        Returns:
            current font weight
        """
        return WEIGHT.inverse[self.weight()]

    def set_capitalization(self, capitalization: CapitalizationStr):
        """Set the font capitalization.

        Args:
            capitalization: font capitalization

        Raises:
            InvalidParamError: invalid font capitalization
        """
        if capitalization not in CAPITALIZATION:
            raise InvalidParamError(capitalization, CAPITALIZATION)
        self.setCapitalization(CAPITALIZATION[capitalization])

    def get_capitalization(self) -> CapitalizationStr:
        """Get current font capitalization.

        Returns:
            current font capitalization
        """
        return CAPITALIZATION.inverse[self.capitalization()]

    def set_hinting_preference(self, preference: HintingPreferenceStr):
        """Set the hinting preference.

        Args:
            preference: hinting preference

        Raises:
            InvalidParamError: invalid hinting preference
        """
        if preference not in HINTING_PREFERENCE:
            raise InvalidParamError(preference, HINTING_PREFERENCE)
        self.setHintingPreference(HINTING_PREFERENCE[preference])

    def get_hinting_preference(self) -> HintingPreferenceStr:
        """Get current hinting preference.

        Returns:
            current hinting preference
        """
        return HINTING_PREFERENCE.inverse[self.hintingPreference()]

    def set_letter_spacing(self, typ: SpacingTypeStr, spacing: float):
        """Set the letter spacing.

        Args:
            typ: letter spacing type
            spacing: spacing

        Raises:
            InvalidParamError: invalid letter spacing type
        """
        if typ not in SPACING_TYPE:
            raise InvalidParamError(typ, SPACING_TYPE)
        self.setLetterSpacing(SPACING_TYPE[typ], spacing)

    def get_letter_spacing_type(self) -> SpacingTypeStr:
        """Get current letter spacing type.

        Returns:
            current letter spacing type
        """
        return SPACING_TYPE.inverse[self.letterSpacingType()]

    def set_style(self, style: StyleStr):
        """Set the font style.

        Args:
            style: font style

        Raises:
            InvalidParamError: invalid font style
        """
        if style not in STYLE:
            raise InvalidParamError(style, STYLE)
        self.setStyle(STYLE[style])

    def get_style(self) -> StyleStr:
        """Get current font style.

        Returns:
            current font style
        """
        return STYLE.inverse[self.style()]


if __name__ == "__main__":
    font = Font("Consolas")



from qtpy.QtCharts import QtCharts

from prettyqt import core


QtCharts.QAbstractSeries.__bases__ = (core.Object,)


class AbstractSeries(QtCharts.QAbstractSeries):
    """QAbstractSeries with some custom properties."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._group = ""

    def get_group(self):
        return self._group

    def set_group(self, value):
        self._group = value

    group = core.Property(str, get_group, set_group)



from __future__ import annotations

from importlib import metadata
import pkgutil
from typing import List, Optional, Sequence

from qtpy import QtCore

from prettyqt import constants, core


def load_dist_info(name: str) -> Optional[metadata.Distribution]:
    try:
        return metadata.distribution(name)
    except metadata.PackageNotFoundError:
        return None


def list_system_modules() -> List[metadata.Distribution]:
    modules = (test.name for test in pkgutil.iter_modules())
    distributions = (load_dist_info(i) for i in modules)
    return [i for i in distributions if i is not None]


def list_package_requirements(package_name: str) -> List[metadata.Distribution]:
    dist = metadata.distribution(package_name)
    modules = (i.split(" ")[0] for i in dist.requires) if dist.requires else []
    distributions = (load_dist_info(i) for i in modules)
    return [i for i in distributions if i is not None]


class ImportlibDistributionModel(core.AbstractTableModel):

    HEADER = ["Name", "Version", "Summary", "Homepage", "Author", "License"]

    def __init__(
        self,
        distributions: Sequence[metadata.Distribution],
        parent: Optional[QtCore.QObject] = None,
    ):
        super().__init__(parent)
        self.distributions = distributions

    def rowCount(self, parent=core.ModelIndex()):
        return len(self.distributions) if not parent.isValid() else 0

    def columnCount(self, parent=core.ModelIndex()):
        return len(self.HEADER) if not parent.isValid() else 0

    def headerData(self, offset: int, orientation, role):  # type: ignore
        if role == constants.DISPLAY_ROLE:
            if orientation == constants.HORIZONTAL:
                return self.HEADER[offset]

    def data(self, index, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        if role == constants.DISPLAY_ROLE:
            if index.column() == 0:
                dist = self.distributions[index.row()]
                return dist.metadata["Name"]
            elif index.column() == 1:
                dist = self.distributions[index.row()]
                return dist.version
            elif index.column() == 2:
                dist = self.distributions[index.row()]
                return dist.metadata["Summary"]
            elif index.column() == 3:
                dist = self.distributions[index.row()]
                return dist.metadata["Home-Page"]
            elif index.column() == 4:
                dist = self.distributions[index.row()]
                return dist.metadata["Author"]
            elif index.column() == 5:
                dist = self.distributions[index.row()]
                return dist.metadata["License"]
        elif role == constants.USER_ROLE:
            dist = self.distributions[index.row()]
            return dist

    @classmethod
    def from_system(
        cls, parent: Optional[QtCore.QObject] = None
    ) -> ImportlibDistributionModel:
        distributions = list_system_modules()
        return cls(distributions, parent)

    @classmethod
    def from_package(
        cls, package_name: str, parent: Optional[QtCore.QObject] = None
    ) -> ImportlibDistributionModel:
        distributions = list_package_requirements(package_name)
        return cls(distributions, parent)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    modules = list_system_modules()
    tableview = widgets.TableView()
    model = ImportlibDistributionModel.from_package("prettyqt")
    tableview.set_model(model)
    tableview.show()
    app.main_loop()



from typing import Dict, Literal, Optional, Union

from qtpy import QtCore, QtGui, QtWidgets

from prettyqt import core, gui, widgets
from prettyqt.utils import InvalidParamError, autoslot, bidict


TAB_SHAPES = bidict(
    rounded=QtWidgets.QTabWidget.Rounded, triangular=QtWidgets.QTabWidget.Triangular
)

TabShapeStr = Literal["rounded", "triangular"]

TAB_POSITION = bidict(
    north=QtWidgets.QTabWidget.North,
    south=QtWidgets.QTabWidget.South,
    west=QtWidgets.QTabWidget.West,
    east=QtWidgets.QTabWidget.East,
)

TabPositionStr = Literal["north", "south", "west", "east"]

QtWidgets.QTabWidget.__bases__ = (widgets.Widget,)


class TabWidget(QtWidgets.QTabWidget):
    """Widget for managing the tabs section."""

    def __init__(
        self,
        parent: Optional[QtWidgets.QWidget] = None,
        closable: bool = False,
        detachable: bool = False,
    ) -> None:

        # Basic initalization
        super().__init__(parent)
        self.tabCloseRequested.connect(self.remove_tab)
        self.tab_bar = widgets.TabBar(self)

        self.setTabBar(self.tab_bar)

        # Used to keep a reference to detached tabs since their QMainWindow
        # does not have a parent
        self.detached_tabs: Dict[str, DetachedTab] = dict()
        if detachable:
            self.set_detachable()
        self.set_closable(closable)

    def __len__(self) -> int:
        return self.count()

    def __getitem__(self, index: int) -> QtWidgets.QWidget:
        if isinstance(index, int):
            return self.widget(index)
        else:
            result = self.findChild(QtWidgets.QWidget, index)
            if result is None:
                raise KeyError("Widget not found")
            return result

    def serialize_fields(self):
        return dict(
            tabbar=self.tabBar(),
            widgets=self.get_children(),
            movable=self.isMovable(),
            document_mode=self.documentMode(),
            current_index=self.currentIndex(),
            tab_shape=self.get_tab_shape(),
            # elide_mode=self.get_elide_mode(),
            icon_size=self.iconSize(),
            tab_position=self.get_tab_position(),
        )

    def __setstate__(self, state):
        self.setTabBar(state["tabbar"])
        self.setDocumentMode(state.get("document_mode", False))
        self.setMovable(state.get("movable", False))
        self.set_tab_shape(state.get("tab_shape", "rounded"))
        self.setIconSize(state["icon_size"])
        for (widget, name, icon, tooltip, whatsthis) in state["widgets"]:
            i = self.add_tab(widget, name, icon)
            self.setTabToolTip(i, tooltip)
            self.setTabWhatsThis(i, whatsthis)
        self.setCurrentIndex(state.get("index", 0))

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def set_document_mode(self, state: bool = True) -> None:
        self.setDocumentMode(state)

    def set_tab_shape(self, shape: TabShapeStr) -> None:
        """Set tab shape for the tabwidget.

        Args:
            shape: tab shape to use

        Raises:
            InvalidParamError: tab shape does not exist
        """
        if shape not in TAB_SHAPES:
            raise InvalidParamError(shape, TAB_SHAPES)
        self.setTabShape(TAB_SHAPES[shape])

    def get_tab_shape(self) -> TabShapeStr:
        """Return tab shape.

        Returns:
            tab shape
        """
        return TAB_SHAPES.inverse[self.tabShape()]

    def set_tab_position(self, position: TabPositionStr) -> None:
        """Set tab position for the tabwidget.

        Args:
            position: tab position to use

        Raises:
            InvalidParamError: tab position does not exist
        """
        if position not in TAB_POSITION:
            raise InvalidParamError(position, TAB_POSITION)
        self.setTabPosition(TAB_POSITION[position])

    def get_tab_position(self) -> TabPositionStr:
        """Return tab position.

        Returns:
            tab position
        """
        return TAB_POSITION.inverse[self.tabPosition()]

    def get_children(self) -> list:
        return [
            (
                self.widget(i),
                self.tabText(i),
                self.tab_icon(i),
                self.tabToolTip(i),
                self.tabWhatsThis(i),
            )
            for i in range(self.count())
        ]

    def tab_icon(self, i: int) -> gui.Icon:
        return gui.Icon(self.tabIcon(i))

    def set_detachable(self) -> None:
        self.tab_bar.on_detach.connect(self.detach_tab)
        widgets.Application.call_on_exit(self.close_detached_tabs)
        self.setMovable(True)

    def set_closable(self, closable: bool = True) -> None:
        self.setTabsClosable(closable)

    @autoslot.AutoSlot
    def detach_tab(self, index: int, point: QtCore.QPoint) -> None:
        """Detach tab by removing its contents and placing them in a DetachedTab window.

        Args:
            index (int): index location of the tab to be detached
            point (QtCore.QPoint): screen pos for creating the new DetachedTab window

        Returns:
            None: Description
        """
        # Get the tab content
        name = self.tabText(index)
        icon = self.tabIcon(index)
        if icon.isNull():
            icon = self.window().windowIcon()
        widget = self.widget(index)

        try:
            widget_rect = widget.frameGeometry()
        except AttributeError:
            return

        # Create a new detached tab window
        detached_tab = DetachedTab(name, widget)
        detached_tab.set_modality("none")
        detached_tab.set_icon(icon)
        detached_tab.setGeometry(widget_rect)
        detached_tab.on_close.connect(self.attach_tab)
        detached_tab.move(point)
        detached_tab.show()

        # Create a reference to maintain access to the detached tab
        self.detached_tabs[name] = detached_tab

    def add_tab(
        self,
        item: Union[QtWidgets.QWidget, QtWidgets.QLayout],
        label: str,
        icon: gui.icon.IconType = None,
        position: Optional[int] = None,
        show: bool = False,
    ) -> int:
        if isinstance(item, QtWidgets.QLayout):
            widget = widgets.Widget()
            widget.set_layout(item)
        else:
            widget = item
        if position is None:
            position = len(self)
        if not icon:
            index = self.insertTab(position, widget, label)
        else:
            icon = gui.icon.get_icon(icon)
            index = self.insertTab(position, widget, icon, label)
        if show:
            self.setCurrentIndex(index)
        return index

    def attach_tab(
        self,
        widget: Union[QtWidgets.QWidget, QtWidgets.QLayout],
        name: str,
        icon: gui.icon.IconType = None,
        insert_at: Optional[int] = None,
    ):
        """Re-attach tab.

        Re-attach the tab by removing the content from the DetachedTab window,
        closing it, and placing the content back into the DetachableTabWidget.

        Args:
            widget (Union[QtWidgets.QWidget, QtWidgets.QLayout]): the content widget
                from the DetachedTab window
            name (str): the name of the detached tab
            icon (gui.icon.IconType, optional): the window icon for the detached tab
            insert_at (Optional[int], optional): insert the re-attached tab at the
                given index
        """
        widget.setParent(self)

        # Remove the reference
        del self.detached_tabs[name]

        # Determine if the given image and the main window icon are the same.
        # If they are, then do not add the icon to the tab
        self.add_tab(widget, name, icon=icon, position=insert_at, show=True)

    def close_detached_tabs(self) -> None:
        """Close all tabs that are currently detached."""
        tabs = list(self.detached_tabs.values())
        for tab in tabs:
            tab.close()

    @autoslot.AutoSlot
    def remove_tab(self, index: int) -> None:
        widget = self.widget(index)
        self.removeTab(index)
        if widget is not None:
            widget.deleteLater()

    @autoslot.AutoSlot
    def open_widget(self, widget: QtWidgets.QWidget, title: str = "Unnamed"):
        """Create a tab containing delivered widget."""
        self.add_tab(widget, title, icon="mdi.widgets", show=True)

    def set_tab(
        self, index: int, position: str, widget: Optional[QtWidgets.QWidget] = None
    ) -> None:
        self.tabBar().set_tab(index, position, widget)


class DetachedTab(widgets.MainWindow):
    """Window containing a detached tab.

    When a tab is detached, the contents are placed into this QMainWindow.
    The tab can be re-attached by closing the dialog

    Attributes:
        on_close: signal, emitted when window is closed (widget, title, icon)
    """

    on_close = core.Signal(QtWidgets.QWidget, str, QtGui.QIcon)

    def __init__(self, name, widget):
        super().__init__(None)

        self.set_id(name)
        self.set_title(name)

        self.widget = widget
        self.setCentralWidget(self.widget)
        self.widget.show()

    #  If the window is closed, emit the on_close and give the
    #  content widget back to the DetachableTabWidget
    def closeEvent(self, event):
        self.on_close.emit(self.widget, self.get_id(), self.windowIcon())


if __name__ == "__main__":
    app = widgets.app()
    tab_widget = TabWidget()
    widget = widgets.Widget()
    tab_widget.add_tab(widget, "Test")
    widget_2 = widgets.Widget()
    tab_widget.add_tab(widget_2, "Test 2")
    tab_widget.show()
    app.main_loop()



from qtpy import QtGui

from prettyqt import constants, core, gui
from prettyqt.utils import InvalidParamError, colors


class Pen(QtGui.QPen):
    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __repr__(self):
        return f"{type(self).__name__}({self.get_color()})"

    def set_color(self, color: colors.ColorType):
        color = colors.get_color(color)
        self.setColor(color)

    def get_color(self) -> gui.Color:
        return gui.Color(self.color())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())

    def set_cap_style(self, style: constants.CapStyleStr):
        """Set cap style to use.

        Args:
            style: cap style to use

        Raises:
            InvalidParamError: cap style does not exist
        """
        if style not in constants.CAP_STYLE:
            raise InvalidParamError(style, constants.CAP_STYLE)
        self.setCapStyle(constants.CAP_STYLE[style])

    def get_cap_style(self) -> constants.CapStyleStr:
        """Return current cap style.

        Returns:
            cap style
        """
        return constants.CAP_STYLE.inverse[self.capStyle()]

    def set_join_style(self, style: constants.JoinStyleStr):
        """Set join style to use.

        Args:
            style: join style to use

        Raises:
            InvalidParamError: join style does not exist
        """
        if style not in constants.JOIN_STYLE:
            raise InvalidParamError(style, constants.JOIN_STYLE)
        self.setJoinStyle(constants.JOIN_STYLE[style])

    def get_join_style(self) -> constants.JoinStyleStr:
        """Return current join style.

        Returns:
            join style
        """
        return constants.JOIN_STYLE.inverse[self.joinStyle()]

    def set_style(self, style: constants.PenStyleStr):
        """Set pen style to use.

        Args:
            style: pen style to use

        Raises:
            InvalidParamError: pen style does not exist
        """
        if style not in constants.PEN_STYLE:
            raise InvalidParamError(style, constants.PEN_STYLE)
        self.setStyle(constants.PEN_STYLE[style])

    def get_style(self) -> constants.PenStyleStr:
        """Return current pen style.

        Returns:
            pen style
        """
        return constants.PEN_STYLE.inverse[self.style()]



import datetime
from typing import Union

from qtpy import QtCore

from prettyqt import constants, core
from prettyqt.utils import InvalidParamError, to_datetime


class DateTime(QtCore.QDateTime):
    def __repr__(self):
        template = super().__repr__().split("(")[1]
        return f"{type(self).__name__}({template}"

    def __str__(self):
        return self.toString("yyyy-MM-dd hh:mm:ss.zzzzzz")

    def __reduce__(self):
        return self.__class__, (self.date(), self.time())

    def get_value(self) -> datetime.datetime:
        return to_datetime(self)

    def get_date(self) -> core.Date:
        return core.Date(self.date())

    def get_time(self) -> core.Time:
        return core.Time(self.time())

    def get_timezone(self) -> core.TimeZone:
        return core.TimeZone(self.timeZone())

    def set_timezone(self, zone: Union[str, QtCore.QTimeZone]):
        if isinstance(zone, str):
            self.setTimeZone(core.TimeZone(zone))
        else:
            self.setTimeZone(zone)

    def set_time_spec(self, spec: constants.TimeSpecStr):
        """Set the time specification.

        Args:
            spec: time specification to use

        Raises:
            InvalidParamError: time specification does not exist
        """
        if spec not in constants.TIME_SPEC:
            raise InvalidParamError(spec, constants.TIME_SPEC)
        self.setTimeSpec(constants.TIME_SPEC[spec])

    def get_time_spec(self) -> constants.TimeSpecStr:
        """Return current time specification.

        Returns:
            time specification
        """
        return constants.TIME_SPEC.inverse[self.timeSpec()]

    def to_format(self, fmt: constants.DateFormatStr):
        return self.toString(constants.DATE_FORMAT[fmt])


if __name__ == "__main__":
    date = DateTime(2000, 11, 11, 1, 1)
    dt = DateTime(date)
    print(dt.to_format("iso"))
    print(dt.get_timezone())
    pdt = dt.get_value()
    print(pdt.isoformat())
    print(pdt.tzinfo)



from typing import Literal

from qtpy import QtGui

from prettyqt.utils import InvalidParamError, bidict


METRICS = bidict(
    width=QtGui.QPaintDevice.PdmWidth,
    height=QtGui.QPaintDevice.PdmHeight,
    width_mm=QtGui.QPaintDevice.PdmWidthMM,
    height_mm=QtGui.QPaintDevice.PdmHeightMM,
    num_colors=QtGui.QPaintDevice.PdmNumColors,
    depth=QtGui.QPaintDevice.PdmDepth,
    dpi_x=QtGui.QPaintDevice.PdmDpiX,
    dpi_y=QtGui.QPaintDevice.PdmDpiY,
    physical_dpi_x=QtGui.QPaintDevice.PdmPhysicalDpiX,
    physical_dpi_y=QtGui.QPaintDevice.PdmPhysicalDpiY,
    pixel_ratio=QtGui.QPaintDevice.PdmDevicePixelRatio,
    pixel_ratio_scaled=QtGui.QPaintDevice.PdmDevicePixelRatioScaled,
)

MetricStr = Literal[
    "width",
    "height",
    "width_mm",
    "height_mm",
    "num_colors",
    "depth",
    "dpi_x",
    "dpi_y",
    "physical_dpi_x",
    "physical_dpi_y",
    "pixel_ratio",
    "pixel_ratio_scaled",
]


class PaintDevice(QtGui.QPaintDevice):
    def get_metric(self, metric: MetricStr) -> int:
        """Return metric information.

        Args:
            metric: metric information to get

        Returns:
            metric information
        """
        if metric not in METRICS:
            raise InvalidParamError(metric, METRICS)
        return self.metric(METRICS[metric])


if __name__ == "__main__":
    device = PaintDevice()



from qtpy import QtMultimediaWidgets

from prettyqt import multimedia


QtMultimediaWidgets.QVideoWidgetControl.__bases__ = (multimedia.MediaControl,)


class VideoWidgetControl(QtMultimediaWidgets.QVideoWidgetControl):
    pass



from __future__ import annotations

import logging
from typing import Optional, Union

from qtpy import QtQml

from prettyqt import qml
from prettyqt.utils import bidict


logger = logging.getLogger()

ERROR_TYPES = bidict(
    generic=QtQml.QJSValue.GenericError,
    range=QtQml.QJSValue.RangeError,
    reference=QtQml.QJSValue.ReferenceError,
    syntax=QtQml.QJSValue.SyntaxError,
    type=QtQml.QJSValue.TypeError,
    uri=QtQml.QJSValue.URIError,
)


class JSValue(QtQml.QJSValue):
    def __repr__(self):
        return f"{type(self).__name__}({self.toVariant()})"

    def __len__(self):
        return self.property("length").toVariant()

    def __getitem__(self, index: Union[int, str]):
        return self.property(index).toVariant()

    def __delitem__(self, index: str):
        self.deleteProperty(index)

    def __setitem__(self, index: Union[int, str], value):
        self.setProperty(index, value)

    def __iter__(self):
        iterator = qml.JSValueIterator(self)
        return iter([val for val in iterator])

    def __contains__(self, index: str):
        return self.hasProperty(index)

    def get_value(self):
        return self.toVariant()

    def get_error_type(self) -> Optional[str]:
        error_type = self.errorType()
        return ERROR_TYPES.inverse.get(error_type)

    @classmethod
    def from_object(cls, obj, jsengine) -> JSValue:
        """Convert any python object into a QJSValue (must happen in GUI thread)."""
        if obj is None:
            return cls()
        elif isinstance(obj, list) or isinstance(obj, tuple):
            length = len(obj)
            array = JSValue(jsengine.newArray(length))
            for i, v in enumerate(obj):
                array.setProperty(i, cls.from_object(v, jsengine))
            return array
        elif isinstance(obj, dict):
            array = JSValue(jsengine.newArray())
            for k, v in obj.items():
                array.setProperty(k, cls.from_object(v, jsengine))
            return array
        else:
            try:
                return cls(obj)
            except TypeError:
                logger.debug("unknown type: " + str(obj))
                return cls()


if __name__ == "__main__":
    val = JSValue()
